from nuclia_eval.metrics.base import DiscreteScoreResponse, Metric

GROUNDEDNESS_TEMPLATE = """\
You are an INFORMATION OVERLAP classifier. Your task is to determine the extent of overlap between the information in the STATEMENT and the SOURCE.

Information overlap is defined as the degree to which the STATEMENT contains information that is substantially similar or identical to that in the SOURCE. When evaluating overlap, differences in language, phrasing, or structure should not be considered.

Respond by reporting the groundedness metric with the provided function

Groudedness scoring guidelines, the score values range from 0 to 5:

- SCORE 0: No information overlap
- SCORE 1: Minimal information overlap
- SCORE 2: Some information overlap
- SCORE 3: Moderate information overlap
- SCORE 4: Extensive information overlap
- SCORE 5: Complete information overlap


STATEMENT: 
```
\"""
{answer}
\"""
```

SOURCE:
```
\"""
{context}
\"""
```

GROUNDEDNESS SCORE: """

GROUNDEDNESS_TOOL = {
    "type": "function",
    "function": {
        "name": "groundedness",
        "description": "Is the answer grounded in any of the provided contexts?",
        "parameters": {
            "type": "object",
            "properties": DiscreteScoreResponse.model_json_schema()["properties"],
            "required": DiscreteScoreResponse.model_json_schema()["required"],
        },
    },
}

Groundedness = Metric(
    template=GROUNDEDNESS_TEMPLATE,
    response_model=DiscreteScoreResponse,
    tool=GROUNDEDNESS_TOOL,
)



from pathlib import Path

from pydantic import Field
from pydantic_settings import BaseSettings


class Settings(BaseSettings):
    """General Settings for nuclia-eval, any of these settings can be overridden by providing a custom Settings object to the evaluator or by setting the environment variable with the same name."""

    nuclia_model_cache: str = Field(
        default=str(Path.home().joinpath(".nuclia-model-cache")),
        description="The path to the model cache directory.",
    )



from nuclia_eval.metrics.base import DiscreteScoreResponse, Metric

CONTEXT_RELEVANCE_TEMPLATE = """\
You are a RELEVANCE grader, tasked with providing relevance scores between a given QUESTION and the CONTEXT provided.
Respond by reporting the context relevance metric with the provided function, where the score value ranges from 0 (no relevance) to 5 (entirely relevant).

Additional Scoring Guidelines:

- Long and short CONTEXTS should be equally considered for relevance assessment.
- Language differences should not influence the score.
- The relevance score should increase as the CONTEXT provides more relevant information to the QUESTION.
- Higher scores indicate relevance to more parts of the QUESTION.
- A score of 1 indicates relevance to some parts, while 2 or 3 suggests relevance to most parts.
- Scores of 4 or 5 should be reserved for CONTEXT that is relevant to the entire QUESTION, with higher scores indicating greater relevance.
- CONTEXT must be helpful for answering the entire QUESTION to receive a score of 5.


QUESTION:
```
\"""
{query}
\"""
```

CONTEXT:
```
\"""
{context}
\"""
```

CONTEXT RELEVANCE SCORE: """

CONTEXT_RELEVANCE_TOOL = {
    "type": "function",
    "function": {
        "name": "context_relevance",
        "description": "Is the retrieved context relevant to the question?",
        "parameters": {
            "type": "object",
            "properties": DiscreteScoreResponse.model_json_schema()["properties"],
            "required": DiscreteScoreResponse.model_json_schema()["required"],
        },
    },
}

ContextRelevance = Metric(
    template=CONTEXT_RELEVANCE_TEMPLATE,
    response_model=DiscreteScoreResponse,
    tool=CONTEXT_RELEVANCE_TOOL,
)



from pathlib import Path
from typing import Union

import safetensors.torch
from torch import nn


def load_lora_low_mem(
    model: nn.Module, lora_path: Union[Path, str], scaling: float = 2.0
):  # pragma: no cover
    """Loads LoRA checkpoint with a low memory footprint compared to the official implementation"""
    lora_path = Path(lora_path)
    assert lora_path.is_file(), f"{lora_path} does not exist or is not a file"

    lora_state_dict = safetensors.torch.load_file(lora_path)

    lora_dtypes = set([p.dtype for p in lora_state_dict.values()])
    assert (
        len(lora_dtypes) == 1
    ), f"LoRA weights have multipe different dtypes {lora_dtypes}. All weights need to have the same dtype"
    lora_dtype = lora_dtypes.pop()
    assert (
        lora_dtype == model.dtype
    ), f"LoRA weights dtype differs from model's dtype {lora_dtype} != {model.dtype}"
    assert all("lora" in key for key in lora_state_dict.keys())

    state_dict = model.state_dict()

    if model.args.lora is None:
        # move tensors to device
        lora_state_dict = {k: v.to(model.device) for k, v in lora_state_dict.items()}
        named_modules = dict(model.named_modules())
        for name, module in named_modules.items():
            if isinstance(module, nn.Linear) and name != "output":
                layer_id = name.split(".")[1]
                if layer_id in model.layers:
                    weight = (
                        module.weight
                        + (
                            lora_state_dict[name + ".lora_B.weight"]
                            @ lora_state_dict[name + ".lora_A.weight"]
                        )
                        * scaling
                    )
                    state_dict[name + ".weight"].copy_(weight)
    else:
        for k, v in lora_state_dict.items():
            state_dict.update(lora_state_dict)

            layer_id = k.split(".")[1]
            if layer_id in model.layers:
                state_dict[k] = v

    model.load_state_dict(state_dict, strict=True)

    # Clear any remaining variables to free up memory
    del lora_state_dict
    del state_dict


def inherit_docstrings(cls):  # pragma: no cover
    for name, func in vars(cls).items():
        if not func.__doc__ and hasattr(getattr(cls, name), "__doc__"):
            func.__doc__ = getattr(cls, name).__doc__
    return cls



from abc import ABC, abstractmethod
from typing import Tuple

from nuclia_eval.metrics.base import DiscreteScoreReasonResponse, DiscreteScoreResponse


class RAGEvaluator(ABC):  # pragma: no cover
    """Base class for all RAG evaluators"""

    @abstractmethod
    def evaluate_rag(
        self, query: str, answer: str, contexts: list[str]
    ) -> Tuple[
        DiscreteScoreReasonResponse,
        list[DiscreteScoreResponse],
        list[DiscreteScoreResponse],
    ]:
        """This method evaluates a whole RAG experience, given the user's query, the model's answer and contexts retrieved at the retrieval phase it computes the answer relevance, context relevance and groundedness.

        * Answer relevance refers to the directness and appropriateness of the response in addressing the specific question asked, providing accurate, complete, and contextually suitable information.
        * The context relevance is the relevance of the **context** to the **question**, on a scale of 0 to 5.
        * Groundedness is defined as the degree of information overlap to which the **answer** contains information that is substantially similar or identical to that in the **context** piece. The scores are between 0 and 5.


        For more information on the RAG evaluation, you can see each metric's definition at nuclia-eval/src/nuclia_eval/metrics/

        Args:
            query (str): The user's query
            answer (str): The model's answer to the user's query
            contexts (list[str]): The contexts retrieved at the retrieval phase that were used to generate the answer

        Returns:
            Tuple[ DiscreteScoreReasonResponse, list[DiscreteScoreResponse], list[DiscreteScoreResponse], ]: A tuple containing the evaluation result for the answer relevance, the context relevance results and groundedness results
        """
        ...

    @abstractmethod
    def answer_relevance(
        self,
        query: str,
        answer: str,
    ) -> DiscreteScoreReasonResponse:
        """This method evaluates the relevance of the model's answer to the user's query.

        Answer relevance refers to the directness and appropriateness of the response in addressing the specific question asked, providing accurate, complete, and contextually suitable information.

        For more information on answer relevance, see the metric's definition at nuclia-eval/src/nuclia_eval/metrics/answer_relevance.py

        Args:
            query (str): The user's query
            answer (str): The model's answer to the user's query

        Returns:
            DiscreteScoreReasonResponse: The evaluation result for the answer relevance
        """
        ...

    @abstractmethod
    def context_relevance(
        self, query: str, contexts: list[str]
    ) -> list[DiscreteScoreResponse]:
        """This method evaluates the relevance of the contexts retrieved at the retrieval phase to the user's query.

        The context relevance is the relevance of the **context** to the **question**, on a scale of 0 to 5.

        For more information on context relevance, see the metric's definition at nuclia-eval/src/nuclia_eval/metrics/context_relevance.py

        Args:
            query (str): The user's query
            contexts (list[str]): The contexts retrieved at the retrieval phase

        Returns:
            list[DiscreteScoreResponse]: The evaluation results for the context relevance
        """
        ...

    @abstractmethod
    def groundedness(
        self, answer: str, contexts: list[str]
    ) -> list[DiscreteScoreResponse]:
        """This method evaluates the groundedness of the model's answer to the contexts retrieved at the retrieval phase.
        Groundedness is defined as the degree of information overlap to which the **answer** contains information that is substantially similar or identical to that in the **context** piece. The scores are between 0 and 5.
        For more information on groundedness, see the metric's definition at nuclia-eval/src/nuclia_eval/metrics/groundedness.py

        Args:
            answer (str): The model's answer to the user's query
            contexts (list[str]): The contexts retrieved at the retrieval phase

        Returns:
            list[DiscreteScoreResponse]: The evaluation results for the groundedness
        """
        ...



"""This module contains definition of the metrics used to evaluate the quality of the generated answers."""

from nuclia_eval.metrics.answer_relevance import AnswerRelevance
from nuclia_eval.metrics.context_relevance import ContextRelevance
from nuclia_eval.metrics.groundedness import Groundedness

__all__ = [
    "AnswerRelevance",
    "ContextRelevance",
    "Groundedness",
]



import json
from pathlib import Path
from typing import Optional, Tuple, Type, TypeVar

import torch
from huggingface_hub import snapshot_download
from mistral_common.protocol.instruct.messages import (
    ChatMessageType,
    SystemMessage,
    UserMessage,
)
from mistral_common.protocol.instruct.request import ChatCompletionRequest
from mistral_common.protocol.instruct.tool_calls import FunctionCall, Tool, ToolChoice
from mistral_common.tokens.tokenizers.mistral import MistralTokenizer
from mistral_inference.generate import generate
from mistral_inference.transformer import Transformer
from pydantic import BaseModel, ValidationError

from nuclia_eval import logger
from nuclia_eval.exceptions import InvalidToolCallException
from nuclia_eval.metrics import (
    AnswerRelevance,
    ContextRelevance,
    Groundedness,
)
from nuclia_eval.metrics.base import (
    DiscreteScoreReasonResponse,
    DiscreteScoreResponse,
    Metric,
)
from nuclia_eval.models.base import RAGEvaluator
from nuclia_eval.settings import Settings
from nuclia_eval.utils import inherit_docstrings, load_lora_low_mem

T = TypeVar("T", bound=BaseModel)


class REMiEvaluator(RAGEvaluator):
    """Evaluator that uses a REMi model"""

    def __init__(
        self,
        settings: Optional[Settings] = None,
        force_download: bool = False,
        device: str = "cuda",
    ) -> None:
        super().__init__()
        # Load default settings if not provided
        if settings is None:
            settings = Settings()
        self.settings = settings

        # Download models
        self._download_base_model(force_download)
        self._download_adapter_model(force_download)

        # Load tokenizer
        logger.info("Loading tokenizer")
        self.tokenizer = MistralTokenizer.from_file(
            str(self._base_model_path / "tokenizer.model.v3")
        )
        logger.info("Tokenizer loaded successfully")

        # Load model
        logger.info("Loading base model")
        self.model = Transformer.from_folder(
            str(self._base_model_path), dtype=torch.float16, device=device
        )
        logger.info("Base model loaded successfully")

        # Load LoRA
        logger.info("Loading REMi adapter model")
        load_lora_low_mem(self.model, self._adapter_model_path / "lora.safetensors")
        logger.info("REMi adapter model loaded successfully")

        # Move model to desired device
        self.device = device
        logger.info(f"Moving model to {self.device}")
        self.model = self.model.to(self.device)
        logger.info(f"Model moved to {self.device}")

        # Cache variables
        self._tools: Optional[list[Tool]] = None
        self._system_message: Optional[SystemMessage] = None

    def _download_base_model(self, force_download: bool = False):
        self._base_model_path = (
            Path(self.settings.nuclia_model_cache) / "Mistral-7B-Instruct-v0.3"
        )
        if force_download or not self._base_model_path.exists():
            logger.info(
                f"Downloading base model to {self._base_model_path}, to override this behavior, please provide the path in the settings or as an environment variable `NUCLIA_MODEL_CACHE`"
            )
            snapshot_download(
                repo_id="mistralai/Mistral-7B-Instruct-v0.3",
                allow_patterns=[
                    "params.json",
                    "consolidated.safetensors",
                    "tokenizer.model.v3",
                ],
                local_dir=self._base_model_path,
            )
            logger.info("Base model downloaded successfully")
        else:
            logger.info(
                f"Base model already exists at {self._base_model_path}, skipping download"
            )

    def _download_adapter_model(self, force_download: bool = False):
        self._adapter_model_path = Path(self.settings.nuclia_model_cache) / "REMi-v0"
        if force_download or not self._adapter_model_path.exists():
            logger.info(
                f"Downloading adapter model to {self._adapter_model_path}, to override this behavior, please provide the path in the settings or as an environment variable `NUCLIA_MODEL_CACHE`"
            )
            snapshot_download(
                repo_id="nuclia/REMi-v0",
                local_dir=self._adapter_model_path,
            )
            logger.info("Adapter model downloaded successfully")
        else:
            logger.info(
                f"Adapter model already exists at {self._adapter_model_path}, skipping download"
            )

    @inherit_docstrings
    def evaluate_rag(
        self, query: str, answer: str, contexts: list[str]
    ) -> Tuple[
        DiscreteScoreReasonResponse,
        list[DiscreteScoreResponse],
        list[DiscreteScoreResponse],
    ]:
        # Answer relevance
        answer_relevance = self.answer_relevance(query, answer)
        # Context relevance
        context_relevance = self.context_relevance(query, contexts)
        # Groundedness
        groundedness = self.groundedness(answer, contexts)
        return answer_relevance, context_relevance, groundedness

    def answer_relevance(self, query: str, answer: str) -> DiscreteScoreReasonResponse:
        system_message = self._get_system_message()
        answer_relevance_message = self._get_metric_message(
            AnswerRelevance, query=query, answer=answer
        )
        response = self._chat_completion_request(
            [system_message, answer_relevance_message],
            Tool.model_validate(AnswerRelevance.tool),
            DiscreteScoreReasonResponse,  # type: ignore
        )
        return response

    @inherit_docstrings
    def groundedness(
        self, answer: str, contexts: list[str]
    ) -> list[DiscreteScoreResponse]:
        system_message = self._get_system_message()
        groundedness_messages = [
            self._get_metric_message(Groundedness, answer=answer, context=context)
            for context in contexts
        ]
        responses = []
        for message in groundedness_messages:
            responses.append(
                self._chat_completion_request(
                    [system_message, message],
                    Tool.model_validate(Groundedness.tool),
                    DiscreteScoreResponse,  # type: ignore
                )
            )
        return responses

    @inherit_docstrings
    def context_relevance(
        self, query: str, contexts: list[str]
    ) -> list[DiscreteScoreResponse]:
        system_message = self._get_system_message()
        context_relevance_messages = [
            self._get_metric_message(ContextRelevance, query=query, context=context)
            for context in contexts
        ]
        responses = []
        for message in context_relevance_messages:
            responses.append(
                self._chat_completion_request(
                    [system_message, message],
                    Tool.model_validate(ContextRelevance.tool),
                    DiscreteScoreResponse,  # type: ignore
                )
            )
        return responses

    def _chat_completion_request(
        self, messages: list[ChatMessageType], tool: Tool, target_model: Type[T]
    ) -> T:
        request = ChatCompletionRequest(
            messages=messages,
            tools=[tool],
            tool_choice=ToolChoice.any,  # type: ignore
        )
        tokens = self.tokenizer.encode_chat_completion(request).tokens
        out_tokens, _ = generate(
            [tokens],
            self.model,
            max_tokens=512,
            temperature=0.0,
            eos_id=self.tokenizer.instruct_tokenizer.tokenizer.eos_id,
        )
        response = self._validate_generation(
            out_tokens, target_model, tool.function.name
        )
        return response

    def _validate_generation(
        self, out_tokens: list[list[int]], target_model: Type[T], desired_tool_name: str
    ) -> T:
        if not out_tokens or not out_tokens[0]:
            raise InvalidToolCallException("No output generated")
        # First token must be token call
        if out_tokens[0][0] != 5:
            raise InvalidToolCallException("First token is not a tool call")

        result = self.tokenizer.instruct_tokenizer.tokenizer.decode(out_tokens[0])
        try:
            calls = json.loads(result)
            call = FunctionCall.model_validate(calls[0])
            if call.name != desired_tool_name:
                raise InvalidToolCallException("Unexpected tool call")
            model = target_model.model_validate_json(call.arguments)
        except (TypeError, ValueError, KeyError, ValidationError):
            raise InvalidToolCallException("Could not parse response")
        return model

    def _get_system_message(self) -> SystemMessage:
        if self._system_message is None:
            self._system_message = SystemMessage(
                content="You are an AI specialized in computing metrics for evaluating Retrieval Augmented Generation (RAG) experiences, use the tools at your disposal to report each of the metrics requested by the user.",
            )
        return self._system_message

    def _get_metric_message(self, metric: Metric, **template_fields) -> UserMessage:
        return UserMessage(
            content=metric.template.format(**template_fields),
        )



import os
from time import monotonic
from unittest.mock import ANY, MagicMock, call, patch

import pytest

from nuclia_eval import REMi
from nuclia_eval.exceptions import InvalidToolCallException
from nuclia_eval.settings import Settings

MANUAL_TEST = os.getenv("MANUAL_TEST", False)


@patch("nuclia_eval.models.remi.generate")
@patch("nuclia_eval.models.remi.snapshot_download")
@patch("nuclia_eval.models.remi.load_lora_low_mem")
@patch("nuclia_eval.models.remi.Transformer")
@patch("nuclia_eval.models.remi.MistralTokenizer")
def test_REMi_evaluator_mock(
    tokenizer_mock: MagicMock,
    transformer_mock: MagicMock,
    lora_load_mock: MagicMock,
    snapshot_download_mock: MagicMock,
    generate_mock: MagicMock,
):
    # Setup mocks
    snapshot_download_mock.return_value = None
    lora_load_mock.return_value = None
    fake_tokenizer = MagicMock()
    tokenizer_mock.from_file.return_value = fake_tokenizer
    fake_model = MagicMock()
    transformer_mock.from_folder.return_value = fake_model

    # Create custom settings
    settings = Settings(
        nuclia_model_cache="my_cache/",
    )
    # Run code
    evaluator = REMi(settings=settings, device="my_device")
    # Check evaluator variables are properly set
    assert evaluator.settings == settings
    # Check that the paths start with the cache path
    assert evaluator._base_model_path.match(settings.nuclia_model_cache + "*")
    assert evaluator._adapter_model_path.match(settings.nuclia_model_cache + "*")

    # Check mock calls
    # 2 Downloads should have been made, one for the base model and one for the adapter model
    snapshot_download_mock.assert_has_calls(
        [
            call(
                repo_id="mistralai/Mistral-7B-Instruct-v0.3",
                allow_patterns=ANY,
                local_dir=evaluator._base_model_path,
            ),
            call(
                repo_id="nuclia/REMi-v0",
                local_dir=evaluator._adapter_model_path,
            ),
        ],
        any_order=True,
    )
    # Check that the tokenizer was loaded
    tokenizer_mock.from_file.assert_called_once()
    # Check that the model was loaded
    transformer_mock.from_folder.assert_called_once()
    # Check that Lora was loaded
    lora_load_mock.assert_called_once()
    # Check that the model was moved to the device
    fake_model.to.assert_called_once_with("my_device")

    # Now load one with default settings
    evaluator = REMi()

    # Configure mocks for a rag_evaluation call
    tokenizer_mock.encode_chat_completion.return_value = [1, 2, 3]
    generate_mock.return_value = ([[5, 123, 123]], [[0.1, 0.2, 0.3]])
    fake_tokenizer.instruct_tokenizer.tokenizer.decode.side_effect = [
        # Answer relevance
        '[{"name": "answer_relevance", "arguments": {"reason": "fake", "score": 1}}]',
        # Context relevances
        '[{"name": "context_relevance", "arguments": {"score": 2}}]',
        '[{"name": "context_relevance", "arguments": {"score": 3}}]',
        # Groundedness
        '[{"name": "groundedness", "arguments": {"score": 4}}]',
        '[{"name": "groundedness", "arguments": {"score": 5}}]',
    ]

    # Run evaluation
    answer_relevance, context_relevances, groundednesses = evaluator.evaluate_rag(
        "query", "answer", ["context1", "context2"]
    )
    assert answer_relevance.score == 1 and answer_relevance.reason == "fake"
    assert [cr.score for cr in context_relevances] == [2, 3]
    assert [g.score for g in groundednesses] == [4, 5]

    # Create another evaluator, so that we can check that the model is not downloaded again
    with patch("pathlib.Path.exists", return_value=True):
        evaluator = REMi(settings=settings, device="my_device")
    # Check that the download calls were not made again
    # assert len(snapshot_download_mock.mock_calls) == 2

    # Check that we raise an error if the first token is not a tool call token
    generate_mock.return_value = ([[123, 123]], [[0.2, 0.3]])
    with pytest.raises(InvalidToolCallException):
        evaluator.evaluate_rag("query", "answer", ["context1", "context2"])

    # Check that we raise an error if the tool call generated is invalid
    generate_mock.return_value = ([[5, 123, 123]], [[0.1, 0.2, 0.3]])
    fake_tokenizer.instruct_tokenizer.tokenizer.decode.side_effect = [
        # Answer relevance with a parameter naming error
        '[{"name": "answer_relevance", "arguments": {"reasoning": "fake", "score": 1}}]',
    ]
    with pytest.raises(InvalidToolCallException):
        evaluator.evaluate_rag("query", "answer", ["context1", "context2"])

    # Check that we raise an error if no output is generated
    generate_mock.return_value = ([], [])
    with pytest.raises(InvalidToolCallException):
        evaluator.evaluate_rag("query", "answer", ["context1", "context2"])

    # Check that we raise an error if the tool name is not the expected one
    generate_mock.return_value = ([[5, 123, 123]], [[0.1, 0.2, 0.3]])
    fake_tokenizer.instruct_tokenizer.tokenizer.decode.side_effect = [
        # Answer relevance but with a different name
        '[{"name": "answer_rel", "arguments": {"reason": "fake", "score": 1}}]',
    ]
    with pytest.raises(InvalidToolCallException):
        evaluator.evaluate_rag("query", "answer", ["context1", "context2"])


@pytest.mark.skipif(
    not MANUAL_TEST,
    reason="This test requires a GPU and the downloaded models and is skipped by default.",
)
def test_REMi_evaluator():
    # Create an instance of the REMiEvaluator class
    t0 = monotonic()
    evaluator = REMi()
    t1 = monotonic()

    query = "By how many Octaves can I shift my OXYGEN PRO 49 keyboard?"
    contexts = [
        """\
* Oxygen Pro 49's keyboard can be shifted 3 octaves down or 4 octaves up.
* Oxygen Pro 61's keyboard can be shifted 3 octaves down or 3 octaves up.

To change the transposition of the keyboard, press and hold Shift, and then use the Key Octave –/+ buttons to lower or raise the keybed by one one, respectively.
The display will temporarily show TRANS and the current transposition (-12 to 12).""",
        """\
To change the octave of the keyboard, use the Key Octave –/+ buttons to lower or raise the octave, respectively
The display will temporarily show OCT and the current octave shift.\n\nOxygen Pro 25's keyboard can be shifted 4 octaves down or 5 octaves up""",
        """\
If your DAW does not automatically configure your Oxygen Pro series keyboard, please follow the setup steps listed in the Oxygen Pro DAW Setup Guides.
To set the keyboard to operate in Preset Mode, press the DAW/Preset Button (on the Oxygen Pro 25) or Preset Button (on the Oxygen Pro 49 and 61).
On the Oxygen Pro 25 the DAW/Preset button LED will be off to show that Preset Mode is selected.
On the Oxygen Pro 49 and 61 the Preset button LED will be lit to show that Preset Mode is selected.""",
    ]
    answer = "Based on the context provided, the Oxygen Pro 49's keyboard can be shifted 3 octaves down or 4 octaves up."

    result = evaluator.evaluate_rag(query=query, answer=answer, contexts=contexts)
    answer_relevance, context_relevances, groundednesses = result
    t2 = monotonic()
    print(
        f"\nAnswer relevance: {answer_relevance.score}, {answer_relevance.reason}"
    )  # 4
    print("Context relevances ", [cr.score for cr in context_relevances])  # [5, 1, 0]
    print("Groundedness: ", [g.score for g in groundednesses])  # [2, 0, 0]
    # Print timings
    print(f"\nTime to load model: {t1 - t0:.2f}s")  # ~15s
    print(f"Time to evaluate: {t2 - t1:.2f}s")  # ~4s

    answer = "Based on the context provided, the Oxygen Pro 61's keyboard can be shifted 4 octaves down or 5 octaves up."
    answer_relevance = evaluator.answer_relevance(query=query, answer=answer)
    context_relevances = evaluator.context_relevance(query=query, contexts=contexts)
    groundednesses = evaluator.groundedness(answer=answer, contexts=contexts)
    t2 = monotonic()
    print(
        f"\nAnswer relevance: {answer_relevance.score}, {answer_relevance.reason}"
    )  # 1
    print("Context relevances ", [cr.score for cr in context_relevances])  # [5, 1, 0]
    print("Groundedness: ", [g.score for g in groundednesses])  # [0, 2, 0]



class ModelException(Exception):
    """Generic exception for model errors."""

    pass


class InvalidToolCallException(ModelException):
    """Exception for when a model does not generate an output that can be mapped to the desired metric."""

    pass



from typing import Any, Type

from pydantic import BaseModel, Field


class Metric(BaseModel):
    """Base class for all metrics"""

    template: str = Field(
        description="The template that will be used to create the user prompt, it can contain {query}, {answer} or/and {context}"
    )
    response_model: Type[BaseModel] = Field(
        description="The output we expect from the model"
    )
    tool: dict[str, Any] = Field(
        description="The tool definition that will given to the model to generate the output"
    )


class DiscreteScoreResponse(BaseModel):
    score: int = Field(
        ge=0, le=5, description="The score of the metric, on a scale of 0 to 5"
    )


class DiscreteScoreReasonResponse(DiscreteScoreResponse):
    reason: str = Field(
        description="The reason for the score, limited to 150 characters"
    )



"""nuclia-eval is a library that simplifies evaluating the RAG experience using nuclia's models."""

import logging

logger = logging.getLogger(__name__)

from nuclia_eval.models.remi import REMiEvaluator as REMi  # noqa: E402

__all__ = ["REMi"]



from nuclia_eval.metrics.base import DiscreteScoreReasonResponse, Metric

ANSWER_RELEVANCE_TEMPLATE = """You are a RELEVANCE grader, tasked with assessing the relevance of a given RESPONSE to a given QUERY and providing a score along with a brief REASON. Relevance refers to the directness and appropriateness of the response in addressing the specific question asked, providing accurate, complete, and contextually suitable information.

Respond by reporting the answer relevance metric with the provided function

Additional scoring guidelines:
- Long and short responses should be scored equally.
- Relevance score should increase as the response provides relevant context to more parts of the query.

- SCORE 0: RESPONSE is relevant to none of the QUERY.
- SCORE 1: RESPONSE is relevant to some parts of the QUERY.
- SCORE 2: RESPONSE is relevant to most parts of the QUERY but contains superfluous information.
- SCORE 3: RESPONSE is relevant to almost all parts of the QUERY or to the entire QUERY but contains superfluous information.
- SCORE 4: RESPONSE is relevant to the entire QUERY.
- SCORE 5: RESPONSE is relevant to the entire QUERY and answers it completely.

The REASON should be brief and clear, explaining why the RESPONSE received the given SCORE. If the SCORE is not a 5, the REASON should contain how the ANSWER could be improved to a 5.


QUERY: {query}

RESPONSE: {answer}

ANSWER RELEVANCE: """

ANSWER_RELEVANCE_TOOL = {
    "type": "function",
    "function": {
        "name": "answer_relevance",
        "description": "The relevance of an answer is its directness and appropriateness in addressing the specific question asked, providing accurate, complete, and contextually suitable information. It ensures clarity and specificity, avoiding extraneous details while fully satisfying the inquiry.",
        "parameters": {
            "type": "object",
            "properties": DiscreteScoreReasonResponse.model_json_schema()["properties"],
            "required": DiscreteScoreReasonResponse.model_json_schema()["required"],
        },
    },
}

AnswerRelevance = Metric(
    template=ANSWER_RELEVANCE_TEMPLATE,
    response_model=DiscreteScoreReasonResponse,
    tool=ANSWER_RELEVANCE_TOOL,
)
